use anyhow::{Context, Result};
use reqwest::{Client, StatusCode};
use serde::{Deserialize, Serialize};
use serde_json::json;
use tracing::{info, error, debug, instrument};

// Department types for organizational structure
#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum Department {
    Legal,
    Finance,
}

// User representation with department info
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct User {
    pub username: String,
    pub department: Department,
}

// VaultClient to interact with Vault API
#[derive(Clone)]
pub struct VaultClient {
    pub client: Client,
    pub addr: String,
    pub token: String,
}

impl VaultClient {
    pub fn new(addr: &str, token: &str) -> Self {
        VaultClient {
            client: Client::new(),
            addr: addr.to_string(),
            token: token.to_string(),
        }
    }

    // Enable required secrets engines
    #[instrument(skip(self))]
    pub async fn setup_secrets_engines(&self) -> Result<()> {
        info!("Setting up Vault secrets engines...");

        // Enable Transit for document signing
        self.enable_secrets_engine("transit", "transit").await?;

        // Enable KV v2 for document storage
        self.enable_secrets_engine("kv-v2", "documents").await?;

        // Enable userpass for authentication
        self.enable_auth_method("userpass").await?;

        info!("Secrets engines setup complete!");
        Ok(())
    }

    // Enable a secrets engine
    #[instrument(skip(self))]
    async fn enable_secrets_engine(&self, engine_type: &str, path: &str) -> Result<()> {
        info!("Enabling {} secrets engine at {}", engine_type, path);

        let url = format!("{}/v1/sys/mounts/{}", self.addr, path);
        let payload = json!({
            "type": engine_type,
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully enabled {} engine at {}", engine_type, path);
                Ok(())
            }
            StatusCode::BAD_REQUEST => {
                // Check if already exists
                let error_text = response.text().await?;
                if error_text.contains("path is already in use") {
                    info!("Secrets engine already enabled at {}", path);
                    Ok(())
                } else {
                    error!("Failed to enable secrets engine: {}", error_text);
                    Err(anyhow::anyhow!("Failed to enable secrets engine: {}", error_text))
                }
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to enable secrets engine: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to enable secrets engine: {} - {}", status, error_text))
            }
        }
    }

    // Enable an auth method
    #[instrument(skip(self))]
    async fn enable_auth_method(&self, method: &str) -> Result<()> {
        info!("Enabling {} auth method", method);

        let url = format!("{}/v1/sys/auth/{}", self.addr, method);
        let payload = json!({
            "type": method,
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully enabled {} auth method", method);
                Ok(())
            }
            StatusCode::BAD_REQUEST => {
                // Check if already exists
                let error_text = response.text().await?;
                if error_text.contains("path is already in use") {
                    info!("Auth method already enabled at {}", method);
                    Ok(())
                } else {
                    error!("Failed to enable auth method: {}", error_text);
                    Err(anyhow::anyhow!("Failed to enable auth method: {}", error_text))
                }
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to enable auth method: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to enable auth method: {} - {}", status, error_text))
            }
        }
    }

    // Create a new user in Vault and associate with department
    #[instrument(skip(self, password))]
    pub async fn create_user(&self, username: &str, password: &str, department: Department) -> Result<()> {
        info!("Creating user {} in department {:?}", username, department);

        // Step 1: Create a policy for the user
        let policy_name = format!("{}-policy", username);
        self.create_signing_policy(&policy_name, department.clone()).await?;

        // Step 2: Create the user with userpass auth
        let url = format!("{}/v1/auth/userpass/users/{}", self.addr, username);
        let payload = json!({
            "password": password,
            "policies": [policy_name],
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully created user {}", username);

                // Step 3: Create a signing key for the user
                self.create_signing_key(username).await?;

                // Step 4: Store user metadata in KV store
                self.store_user_metadata(username, department).await?;

                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to create user: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to create user: {} - {}", status, error_text))
            }
        }
    }

    // Create a signing policy for a user based on their department
    #[instrument(skip(self))]
    async fn create_signing_policy(&self, policy_name: &str, department: Department) -> Result<()> {
        let dept_name = match department {
            Department::Legal => "legal",
            Department::Finance => "finance",
        };

        // Policy content with specific paths for the department
        let policy = format!(r#"
            # Allow reading document metadata
            path "documents/data/docs/*" {{
                capabilities = ["read"]
            }}

            # Allow signing with user's key
            path "transit/sign/{{{{identity.entity.name}}}}" {{
                capabilities = ["update"]
            }}

            # Allow signature verification
            path "transit/verify/*" {{
                capabilities = ["update"]
            }}

            # Department-specific path
            path "documents/data/dept/{}/signatures/*" {{
                capabilities = ["create", "read", "update"]
            }}
        "#, dept_name);

        let url = format!("{}/v1/sys/policies/acl/{}", self.addr, policy_name);
        let payload = json!({
            "policy": policy,
        });

        let response = self.client
            .put(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully created policy {}", policy_name);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to create policy: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to create policy: {} - {}", status, error_text))
            }
        }
    }

    // Create a signing key for a user in the Transit engine
    #[instrument(skip(self))]
    async fn create_signing_key(&self, username: &str) -> Result<()> {
        let url = format!("{}/v1/transit/keys/{}", self.addr, username);
        let payload = json!({
            "type": "ed25519",
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully created signing key for {}", username);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to create signing key: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to create signing key: {} - {}", status, error_text))
            }
        }
    }

    // Store user metadata in the KV store
    #[instrument(skip(self))]
    async fn store_user_metadata(&self, username: &str, department: Department) -> Result<()> {
        let dept_str = match department {
            Department::Legal => "legal",
            Department::Finance => "finance",
        };

        let url = format!("{}/v1/documents/data/users/{}", self.addr, username);
        let payload = json!({
            "data": {
                "username": username,
                "department": dept_str,
            }
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully stored metadata for user {}", username);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to store user metadata: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to store user metadata: {} - {}", status, error_text))
            }
        }
    }

    // Create 10 users with departmental hierarchy - 5 in each department
    #[instrument(skip(self))]
    pub async fn setup_hierarchical_users(&self) -> Result<()> {
        info!("Setting up hierarchical user structure");

        // Create 5 users in Legal department
        for i in 1..=5 {
            let username = format!("legal{}", i);
            let password = format!("legal{}pass", i);
            debug!(username, "Creating Legal department user");
            self.create_user(&username, &password, Department::Legal).await?;
        }

        // Create 5 users in Finance department
        for i in 1..=5 {
            let username = format!("finance{}", i);
            let password = format!("finance{}pass", i);
            debug!(username, "Creating Finance department user");
            self.create_user(&username, &password, Department::Finance).await?;
        }

        // Setup document signing requirements
        self.setup_signing_requirements().await?;

        info!("Successfully created 10 users in hierarchical structure!");
        Ok(())
    }

    // Configure document signing requirements
    #[instrument(skip(self))]
    async fn setup_signing_requirements(&self) -> Result<()> {
        info!("Setting up document signing requirements");

        let url = format!("{}/v1/documents/data/config/signing_requirements", self.addr);
        let payload = json!({
            "data": {
                "total_required": 3,
                "departments": [
                    {
                        "name": "legal",
                        "required": 1,
                        "total_users": 5
                    },
                    {
                        "name": "finance",
                        "required": 1,
                        "total_users": 5
                    }
                ],
                "description": "Requires 3 signatures total, with at least 1 from Legal and 1 from Finance departments"
            }
        });

        let response = self.client
            .post(&url)
            .header("X-Vault-Token", &self.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::NO_CONTENT | StatusCode::OK => {
                info!("Successfully configured signing requirements");
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to configure signing requirements: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to configure signing requirements: {} - {}", status, error_text))
            }
        }
    }

    // Login a user and get their token
    #[instrument(skip(self, password))]
    pub async fn login_user(&self, username: &str, password: &str) -> Result<String> {
        info!("Attempting user login: {}", username);

        let url = format!("{}/v1/auth/userpass/login/{}", self.addr, username);
        let payload = json!({
            "password": password,
        });

        let response = self.client
            .post(&url)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK => {
                let json: serde_json::Value = response.json().await?;
                let token = json["auth"]["client_token"]
                    .as_str()
                    .context("Failed to extract client token")?
                    .to_string();

                info!("User {} successfully logged in", username);
                Ok(token)
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to login: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to login: {} - {}", status, error_text))
            }
        }
    }

    // Get user info including department
    #[instrument(skip(self))]
    pub async fn get_user_info(&self, username: &str) -> Result<User> {
        debug!("Getting user info for: {}", username);

        let url = format!("{}/v1/documents/data/users/{}", self.addr, username);

        let response = self.client
            .get(&url)
            .header("X-Vault-Token", &self.token)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK => {
                let json: serde_json::Value = response.json().await?;
                let department_str = json["data"]["data"]["department"]
                    .as_str()
                    .context("Failed to extract department")?;

                let department = match department_str {
                    "legal" => Department::Legal,
                    "finance" => Department::Finance,
                    _ => {
                        error!("Unknown department: {}", department_str);
                        return Err(anyhow::anyhow!("Unknown department: {}", department_str));
                    }
                };

                debug!("Retrieved user info for {} in {:?} department", username, department);
                Ok(User {
                    username: username.to_string(),
                    department,
                })
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to get user info: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to get user info: {} - {}", status, error_text))
            }
        }
    }
}
